/* CmdBkg (c) 2016-2020 Mikael Sollenborn */

#undef WINVER
#define WINVER 0x0501
#undef _WIN32_WINNT
#define _WIN32_WINNT 0x0501
#define _WIN32_DCOM

// Usage: cmdbkg "file.bmp" [transparency 0-100(default 33) [includeBorders]]
// Specify no arguments to remove previous background.
// Specify /? as first argument to see this help.

// Compilation: gcc -o cmdbkg.exe cmdbkg.c -lgdi32 -ldwmapi -lole32 -lwbemuuid -lgdiplus

// TODO/Issues:
// 1. Allow centered placement of image, no stretching?
// 2. Background window visible a short time after restoring minimized console window. Minimize background if console window is minimized?
// 3. Inelegant polling. Use SetWinEventHook instead?
// (Done, maybe?) 4. Find a better way to kill the old background window than this horrible title polling hack

#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <windows.h>
#include <dwmapi.h>
#include <wbemidl.h>
#include <olectl.h>

// GDI+ headers are C++-only in MS Visual Studio (even the flat API).
// Instead of including <gdiplus.h> we declare the necessary things here. However the gdiplus lib has to be linked.
typedef struct tag_GdiplusStartupInput
{
  UINT32 GdiplusVersion;
  void *DebugEventCallback;
  BOOL SuppressBackgroundThread;
  BOOL SuppressExternalCodecs;
} GdiplusStartupInput;

int __stdcall GdiplusStartup(ULONG_PTR *, const GdiplusStartupInput *, void *);
void __stdcall GdiplusShutdown(ULONG_PTR);
int __stdcall GdipCreateBitmapFromFile(const wchar_t *, void **);
int __stdcall GdipCreateHBITMAPFromBitmap(void *, HBITMAP *, DWORD);

#define POLL_INTERVAL 20
#define SECOND_RUN L"_secondRun_"
#define COLOR_VIRGIN_BKG COLOR_3DSHADOW

// should be available everywhere
HWND glob_hConsoleWnd; // console window handle
LPWSTR *glob_szArgList; // allocated array of command line arguments

// Read the content of an image file, convert it to a bitmap in memory, and return a handle to it.
static HBITMAP GetHBitmapFromFile(wchar_t *szImagePath)
{
  static HBITMAP hBmp = NULL;
  if (hBmp || !szImagePath)
    return hBmp;

  GdiplusStartupInput gpStartupInput = { 1U, NULL, FALSE, FALSE };
  ULONG_PTR gpToken = 0U;
  void *bmpPtr = NULL;
  GdiplusStartup(&gpToken, &gpStartupInput, NULL);
  GdipCreateBitmapFromFile(szImagePath, &bmpPtr);
  if (bmpPtr)
    GdipCreateHBITMAPFromBitmap(bmpPtr, &hBmp, GetSysColor(COLOR_VIRGIN_BKG));

  GdiplusShutdown(gpToken);
  return hBmp;
}

// Load an image file and insert it into a window.
static int LoadImg(HWND hWnd, wchar_t *szImagePath, long x, long y, long z, long w, long h)
{
  HDC hDc = NULL, hDcBmp = NULL;
  HBITMAP hBmp1 = NULL, hBmp2 = NULL;
  HGDIOBJ hGdiObj = NULL;
  BITMAP bmp = { 0 };
  int iRet = EXIT_FAILURE;

  if (hWnd)
  {
    if ((hDc = GetDC(hWnd)))
    {
      if ((hDcBmp = CreateCompatibleDC(hDc)))
      {
        if ((hBmp1 = GetHBitmapFromFile(szImagePath)))
        {
          if (GetObjectW(hBmp1, sizeof(bmp), &bmp))
          {
            if (w == -1)
            {
              if ((w = (long)((double)bmp.bmWidth * z / 100.0 + 0.5)) <= 0 || (h = (long)((double)bmp.bmHeight * z / 100.0 + 0.5)) <= 0)
              {
                w = bmp.bmWidth;
                h = bmp.bmHeight;
              }
            }
            if ((hBmp2 = (HBITMAP)CopyImage((HANDLE)hBmp1, IMAGE_BITMAP, w, h, LR_DEFAULTCOLOR)))
            {
              if ((hGdiObj = SelectObject(hDcBmp, hBmp2)) && hGdiObj != HGDI_ERROR)
              {
                if (BitBlt(hDc, (int)x, (int)y, (int)w, (int)h, hDcBmp, 0, 0, SRCCOPY))
                  iRet = EXIT_SUCCESS;

                DeleteObject(hGdiObj);
              }
              DeleteObject(hBmp2);
            }
          }
        }
        ReleaseDC(hWnd, hDcBmp);
      }
      ReleaseDC(hWnd, hDc);
    }
  }
  return iRet;
}

// Make a console a layered window and let it be tranparent to the specified extend.
static BOOL SetConsoleTransparency(HWND hWnd, long percentage)
{
  BYTE bAlpha = 0;
  LONG_PTR lNewLong = 0L;
  if (hWnd && percentage > -1L && percentage < 101L)
  {
    bAlpha = (BYTE)((255L * (100L - percentage) + 50L) / 100L);
    lNewLong = GetWindowLongPtrW(hWnd, GWL_EXSTYLE) | WS_EX_LAYERED;
    if (!SetWindowLongPtrW(hWnd, GWL_EXSTYLE, lNewLong))
      return FALSE;

    return SetLayeredWindowAttributes(hWnd, 0, bAlpha, LWA_ALPHA);
  }

  return FALSE;
}

// Callback function to find the top level window of a process with the given PID, and send WM_CLOSE to its message queue. (called by EnumWindows)
static BOOL CALLBACK CloseWindowByPid(HWND hwnd, LPARAM lParamPidPtr)
{
  unsigned *pPid = (unsigned *)lParamPidPtr;
  DWORD processId = 0;
  GetWindowThreadProcessId(hwnd, &processId);
  if (*pPid == processId && !GetWindow(hwnd, GW_OWNER))
  {
    SendMessageTimeoutW(hwnd, WM_CLOSE, 0, 0, SMTO_BLOCK | SMTO_NOTIMEOUTIFNOTHUNG, 2000U, NULL);
    return FALSE;
  }

  return TRUE;
}

// Find the PID of the process where the command line contains " _secondRun_ <hwnd>" (<hwnd> = HEX string representing the console window handle).
// Calls EnumWindows using CloseWindowByPid for the found PID.
static void CloseBkgWnd(void)
{
  wchar_t findBuffer[64] = { 0 };
  IWbemLocator *wbemLocator = NULL;
  IWbemServices *wbemServices = NULL;
  IEnumWbemClassObject *enumWbem = NULL;

  wsprintfW(findBuffer, L" " SECOND_RUN L" %p", (void *)glob_hConsoleWnd);

  if (CoInitializeEx(NULL, COINIT_MULTITHREADED) != S_OK ||
      CoInitializeSecurity(NULL, -1, NULL, NULL, RPC_C_AUTHN_LEVEL_DEFAULT, RPC_C_IMP_LEVEL_IMPERSONATE, NULL, EOAC_NONE, NULL) != S_OK ||
      CoCreateInstance(&CLSID_WbemLocator, 0, CLSCTX_INPROC_SERVER, &IID_IWbemLocator, (LPVOID *)&wbemLocator) != S_OK ||
      wbemLocator->lpVtbl->ConnectServer(wbemLocator, L"ROOT\\CIMV2", NULL, NULL, NULL, 0, NULL, NULL, &wbemServices) != S_OK ||
      wbemServices->lpVtbl->ExecQuery(wbemServices, L"WQL", L"SELECT ProcessId,CommandLine FROM Win32_Process", WBEM_FLAG_FORWARD_ONLY, NULL, &enumWbem) != S_OK)
  {
    CoUninitialize();
    LocalFree(glob_szArgList);
    exit((int)GetLastError()); // OK, I was just lazy ...
  }

  if (enumWbem != NULL)
  {
    IWbemClassObject *result = NULL;
    ULONG returnedCount = 0;

    while (enumWbem->lpVtbl->Next(enumWbem, WBEM_INFINITE, 1, &result, &returnedCount) == S_OK)
    {
      VARIANT ProcessId = { 0 };
      VARIANT CommandLine = { 0 };

      result->lpVtbl->Get(result, L"ProcessId", 0, &ProcessId, NULL, NULL);
      result->lpVtbl->Get(result, L"CommandLine", 0, &CommandLine, NULL, NULL);
      if (CommandLine.vt != VT_NULL && wcsstr(CommandLine.bstrVal, findBuffer))
      {
        EnumWindows(CloseWindowByPid, (LPARAM)&ProcessId.uintVal);

        result->lpVtbl->Release(result);
        enumWbem->lpVtbl->Release(enumWbem);
        wbemServices->lpVtbl->Release(wbemServices);
        wbemLocator->lpVtbl->Release(wbemLocator);
        CoUninitialize();
        return;
      }

      result->lpVtbl->Release(result);
    }

    enumWbem->lpVtbl->Release(enumWbem);
  }

  wbemServices->lpVtbl->Release(wbemServices);
  wbemLocator->lpVtbl->Release(wbemLocator);
  CoUninitialize();
}

// Window callback function necessary to handle WM_CLOSE.
static LRESULT CALLBACK WindowProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
  switch (msg)
  {
  case WM_CLOSE:
    SetConsoleTransparency(glob_hConsoleWnd, 0);
    if (glob_szArgList)
    {
      DeleteObject(GetHBitmapFromFile(NULL));
      LocalFree(glob_szArgList);
      glob_szArgList = NULL;
    }

    break;

  default:
    return DefWindowProcW(hwnd, msg, wparam, lparam);
  }

  return 0;
}

int main(void)
{
  const wchar_t scnd[] = L" " SECOND_RUN;
  int bx = 0, by = 0, bw = 0, bh = 0, ret = 0;
  int nbx = 0, nby = 0, nbw = 0, nbh = 0;
  int fw = 0, fh = 0, fth = 0, w = 0, h = 0;
  size_t oldLen = 0;
  HWND hBkgWnd = NULL, currFgWnd = NULL, tempWnd = NULL;
  RECT bounds = { 0 }, extendBounds = { 0 };
  int argCount = 0, initialRedraw = 10;
  long transparency = 33L;
  wchar_t handleBuffer[sizeof(HWND) * 2 + 2] = { 0 }, *oldCmdLine = NULL, *newCmdLine = NULL, *oldCursor = NULL, *newCursor = NULL, *tmpPtr = NULL;
  WNDCLASSW wc = { 0 };
  UINT_PTR timerId = 0;
  BOOL bIsWeirdoBorders = FALSE;
  BOOL res = FALSE;
  MSG msg = { 0 };

  STARTUPINFOW si = { 0 };
  PROCESS_INFORMATION pi = { 0 };

  si.cb = sizeof(si);

  glob_hConsoleWnd = GetConsoleWindow();
  if (!glob_hConsoleWnd)
  {
    MessageBoxW(NULL, L"Cmdbkg error: Unable to prepare window creation (1)", L"Error", MB_OK);
    return 1;
  }

  oldCmdLine = GetCommandLineW();
  glob_szArgList = CommandLineToArgvW(oldCmdLine, &argCount);
  if (glob_szArgList == NULL)
  {
    printf("Cmdbkg error: Unable to get arguments");
    return 1;
  }

  if (argCount == 2 && wcscmp(glob_szArgList[1], L"/?") == 0)
  {
    wprintf(L"\nCmdBkg v1.2 beta : Mikael Sollenborn 2016-2020\n\nUsage: cmdbkg \"file.bmp\" [transparency 0-100(default 33) [includeBorders]]\n\nSpecify no arguments to remove previous background.\n");
    LocalFree(glob_szArgList);
    return 0;
  }

  if (argCount > 1 && wcscmp(glob_szArgList[1], SECOND_RUN) != 0)
  {
    oldLen = wcslen(oldCmdLine);
    newCmdLine = malloc(oldLen * sizeof(wchar_t) + sizeof(scnd) + sizeof(handleBuffer));
    if (!newCmdLine)
    {
      puts("Cmdbkg error: Unable to allocate memory.");
      LocalFree(glob_szArgList);
      return 1;
    }

    // insert two additional argument tokens into the command line, 1st the string "_secondRun_", and 2nd the console window handle as HEX string
    wsprintfW(handleBuffer, L" %p", (void *)glob_hConsoleWnd);
    oldCursor = wcsstr(oldCmdLine, glob_szArgList[0]) + wcslen(glob_szArgList[0]);
    LocalFree(glob_szArgList);
    tmpPtr = wcspbrk(oldCursor, L" \t");
    oldCursor = tmpPtr ? tmpPtr : oldCmdLine + oldLen;
    newCursor = (wchar_t *)memcpy(newCmdLine, oldCmdLine, (size_t)(oldCursor - oldCmdLine) * sizeof(wchar_t)) + (oldCursor - oldCmdLine);
    newCursor = (wchar_t *)memcpy(newCursor, scnd, sizeof(scnd) - sizeof(wchar_t)) + (sizeof(scnd) - sizeof(wchar_t)) / sizeof(wchar_t);
    newCursor = (wchar_t *)memcpy(newCursor, handleBuffer, sizeof(handleBuffer) - sizeof(wchar_t)) + (sizeof(handleBuffer) - sizeof(wchar_t)) / sizeof(wchar_t);
    memcpy(newCursor, oldCursor, (oldLen - (oldCursor - oldCmdLine) + 1) * sizeof(wchar_t));

    // Start the child process.
    if (!CreateProcessW(NULL, // No module name (use command line)
                        newCmdLine, // Command line
                        NULL, // Process handle not inheritable
                        NULL, // Thread handle not inheritable
                        FALSE, // Set handle inheritance to FALSE
                        0, // No creation flags
                        NULL, // Use parent's environment block
                        NULL, // Use parent's starting directory
                        &si, // Pointer to STARTUPINFO structure
                        &pi) // Pointer to PROCESS_INFORMATION structure
    )
    {
      printf("Cmdbkg error: CreateProcess failed (%lu).\n", GetLastError());
      free(newCmdLine);
      return 1;
    }

    CloseHandle(pi.hProcess);
    CloseHandle(pi.hThread);
    free(newCmdLine);
    return 0;
  }

  CloseBkgWnd(); // in case there is already a cmdbkg process running for this console

  GetWindowRect(glob_hConsoleWnd, &bounds);
  bx = bounds.left;
  by = bounds.top;
  bw = bounds.right - bounds.left;
  bh = bounds.bottom - bounds.top;

  FreeConsole();

  SetConsoleTransparency(glob_hConsoleWnd, 0);

  if (argCount < 4)
  {
    LocalFree(glob_szArgList);
    return 0;
  }

  DwmGetWindowAttribute(glob_hConsoleWnd, DWMWA_EXTENDED_FRAME_BOUNDS, &extendBounds, sizeof(extendBounds));
  if (extendBounds.left != bounds.left)
    bIsWeirdoBorders = TRUE;

  if (argCount > 4)
    transparency = wcstol(glob_szArgList[4], NULL, 10);

  if (argCount <= 5 || (argCount > 4 && bIsWeirdoBorders))
  {
    fw = GetSystemMetrics(SM_CXSIZEFRAME);
    if (argCount <= 5)
      fth = GetSystemMetrics(SM_CYSIZEFRAME) + GetSystemMetrics(SM_CYCAPTION);

    fh = fth + GetSystemMetrics(SM_CYSIZEFRAME);
  }

  wc.lpszClassName = L"CmdBkgWindowBackground";
  wc.lpfnWndProc = WindowProc;
  wc.hInstance = NULL; //GetModuleHandle(0);
  wc.hbrBackground = (HBRUSH)(COLOR_VIRGIN_BKG);
  wc.hCursor = LoadCursor(NULL, IDC_ARROW);
  if (!RegisterClassW(&wc))
  {
    MessageBoxW(NULL, L"Cmdbkg error: Unable to prepare window creation (2)", L"Error", MB_OK);
    LocalFree(glob_szArgList);
    return 1;
  }

  hBkgWnd = CreateWindowExW(0, wc.lpszClassName, 0, WS_POPUP | WS_VISIBLE, bx + fw, by + fth, bw - fw * 2, bh - fh, 0, 0, 0, 0);
  if (!hBkgWnd)
  {
    MessageBoxW(NULL, L"Cmdbkg error: Unable to create background window", L"Error", MB_OK);
    LocalFree(glob_szArgList);
    return 1;
  }

  w = bw - fw * 2, h = bh - fth;
  if (w < 1)
    w = 1;

  if (h < 1)
    h = 1;

  ret = LoadImg(hBkgWnd, glob_szArgList[3], 0, 0, 100, w, h);
  if (ret == EXIT_FAILURE)
  {
    MessageBoxW(NULL, L"Cmdbkg error: Unable to load bitmap", L"Error", MB_OK);
    LocalFree(glob_szArgList);
    return 1;
  }

  SetConsoleTransparency(glob_hConsoleWnd, transparency);

  ShowWindow(hBkgWnd, SW_HIDE);
  SetWindowLongPtrW(hBkgWnd, GWL_EXSTYLE, WS_EX_TOOLWINDOW);
  ShowWindow(hBkgWnd, SW_SHOW);

  SetForegroundWindow(glob_hConsoleWnd);
  currFgWnd = GetForegroundWindow();

  while (IsWindow(glob_hConsoleWnd) || IsIconic(glob_hConsoleWnd))
  {
    timerId = SetTimer(NULL, 0, POLL_INTERVAL, NULL);
    res = GetMessage(&msg, 0, 0, 0);
    KillTimer(NULL, timerId);

    if (!(res && msg.message == WM_TIMER && msg.hwnd == NULL && msg.wParam == timerId))
      DispatchMessage(&msg);

    if (!glob_szArgList)
      return 0;

    GetWindowRect(glob_hConsoleWnd, &bounds);
    nbx = bounds.left;
    nby = bounds.top;
    nbw = bounds.right - bounds.left;
    nbh = bounds.bottom - bounds.top;

    tempWnd = GetForegroundWindow();
    if (currFgWnd != tempWnd)
    {
      if (initialRedraw < 5)
        initialRedraw = 5;

      currFgWnd = tempWnd;
    }

    LONG_PTR cStyle = GetWindowLongPtrW(glob_hConsoleWnd, GWL_STYLE);
    if (cStyle & WS_POPUP)
      SetWindowPos(hBkgWnd, glob_hConsoleWnd, nbx, nby, nbw, nbh, SWP_SHOWWINDOW);
    else
      SetWindowPos(hBkgWnd, glob_hConsoleWnd, nbx + fw, nby + fth, nbw - fw * 2, nbh - fh, SWP_SHOWWINDOW);

    if (initialRedraw > 0 || nbx != bx || nby != by || nbw != bw || nbh != bh)
    {
      if (cStyle & WS_POPUP)
        w = nbw, h = nbh;
      else
        w = nbw - fw * 2, h = nbh - fth;

      if (w < 1)
        w = 1;

      if (h < 1)
        h = 1;

      ret = LoadImg(hBkgWnd, glob_szArgList[3], 0, 0, 100, w, h);

      bx = nbx;
      by = nby;
      bw = nbw;
      bh = nbh;
      if (initialRedraw > 0)
        initialRedraw--;
    }
  }

  if (glob_szArgList)
  {
    DeleteObject(GetHBitmapFromFile(NULL));
    LocalFree(glob_szArgList);
    glob_szArgList = NULL;
  }

  return 0;
}
